// -----------------------------------------------------
// Important notice:
// -----------------------------------------------------
//   User Defined List (ttL_myElement)
// -----------------------------------------------------
//   - class ttL_myElement is the list container of the elements (type is myElement), which are the objects
//   - use initLst() in the constructor to set default values to member variables, if any
//   - MARS AUTOMATICALLY calls obLst::readListInput(), which calls processCommand() REPEATLY to process input file
//   - MARS AUTOMATICALLY calls ttL_myElement::initialize() after the input is read,
//   - in ttL_myElement::initialize(), call myElement::init() to initialize the objects as well
//   - put the validity checks in ttL_myElement::initialize()

class myElement : public DeformableTet {
public:
    myElement() { }
    ~myElement() { }
    void init(ReferenceSystem *rs) { DeformableTet::init(rs); }
    void whatAmI() { cout << "myElement" << endl; };
    void reset();
    real calcForces(Node **, real *);
    void setUpStiffnessMatrix(real **);
};

void myElement::reset () {
    if (nsv) {
        for (int i=0; i<nsv; i++)
            stv[i] = 0.;
        if (mat) mat->initHexSV(stv);
    }
}

real myElement::calcForces (Node **ndF, real *f) {
    real c0x = nd[0]->getCx();
    real c0y = nd[0]->getCy();
    real c0z = nd[0]->getCz();
    real v0x = nd[0]->getVx();
    real v0y = nd[0]->getVy();
    real v0z = nd[0]->getVz();
    real c1x = nd[1]->getCx() - c0x;
    real c1y = nd[1]->getCy() - c0y;
    real c1z = nd[1]->getCz() - c0z;
    real v1x = nd[1]->getVx() - v0x;
    real v1y = nd[1]->getVy() - v0y;
    real v1z = nd[1]->getVz() - v0z;
    real c2x = nd[2]->getCx() - c0x;
    real c2y = nd[2]->getCy() - c0y;
    real c2z = nd[2]->getCz() - c0z;
    real v2x = nd[2]->getVx() - v0x;
    real v2y = nd[2]->getVy() - v0y;
    real v2z = nd[2]->getVz() - v0z;
    real c3x = nd[3]->getCx() - c0x;
    real c3y = nd[3]->getCy() - c0y;
    real c3z = nd[3]->getCz() - c0z;
    real v3x = nd[3]->getVx() - v0x;
    real v3y = nd[3]->getVy() - v0y;
    real v3z = nd[3]->getVz() - v0z;
    //  dvi/dxj = vIi dNI/dxj = BIj vIi
    //  xj = xIj NI =>  dxj = xIj dNI  =>  dNI/dxj = | xIj | ^-1
    //                             | c1x  c1y  c1z |
    //  assemble J matrix      J = | c2x  c2y  c2z |
    //                             | c3x  c3y  c3z |
    real A1x = c2y*c3z - c2z*c3y;
    real A1y = c2z*c3x - c2x*c3z;
    real A1z = c2x*c3y - c2y*c3x;
    real A2x = c3y*c1z - c3z*c1y;
    real A2y = c3z*c1x - c3x*c1z;
    real A2z = c3x*c1y - c3y*c1x;
    real A3x = c1y*c2z - c1z*c2y;
    real A3y = c1z*c2x - c1x*c2z;
    real A3z = c1x*c2y - c1y*c2x;
    real dtJ = c1x*A1x + c1y*A1y + c1z*A1z;
    //real vol = fabs(dtJ) / 6.; // volume of Tet element
    //real vol = dtJ / 6.; // volume of Tet element
    //  dtJ /= 6.;
    //  dxi/dxj =  dxj =  dxi = xIi dNI / (xJj dNJ) = B J = I = dij
    //                               | B1x  B2x  B3x |
    //  compute inverse   B = J^-1 = | B1y  B2y  B3y |
    //                               | B1z  B2z  B3z |
    real dJi = 1. / dtJ;
    real B1x = A1x * dJi;
    real B1y = A1y * dJi;
    real B1z = A1z * dJi;
    real B2x = A2x * dJi;
    real B2y = A2y * dJi;
    real B2z = A2z * dJi;
    real B3x = A3x * dJi;
    real B3y = A3y * dJi;
    real B3z = A3z * dJi;
    real dxx = B1x*v1x + B2x*v2x + B3x*v3x;
    real dxy = B1x*v1y + B2x*v2y + B3x*v3y;
    real dxz = B1x*v1z + B2x*v2z + B3x*v3z;
    real dyx = B1y*v1x + B2y*v2x + B3y*v3x;
    real dyy = B1y*v1y + B2y*v2y + B3y*v3y;
    real dyz = B1y*v1z + B2y*v2z + B3y*v3z;
    real dzx = B1z*v1x + B2z*v2x + B3z*v3x;
    real dzy = B1z*v1y + B2z*v2y + B3z*v3y;
    real dzz = B1z*v1z + B2z*v2z + B3z*v3z;

    dyz += dzy;
    dzx += dxz;
    dxy += dyx;
    dyz *= 0.5;
    dzx *= 0.5;
    dxy *= 0.5;

    real ttv[6];
    ttv[0] = dxx;
    ttv[1] = dyy;
    ttv[2] = dzz;
    ttv[3] = dxy;
    ttv[4] = dyz;
    ttv[5] = dzx;

    mat->calcStrsHx(ttv, stv);

    dtJ /= 6.;
    real sxx = - stv[0] * dtJ;
    real syy = - stv[1] * dtJ;
    real szz = - stv[2] * dtJ;
    real sxy = - stv[3] * dtJ;
    real syz = - stv[4] * dtJ;
    real szx = - stv[5] * dtJ;	

    real f1x = B1x*sxx + B1y*sxy + B1z*szx;
    real f1y = B1x*sxy + B1y*syy + B1z*syz;
    real f1z = B1x*szx + B1y*syz + B1z*szz;
    real f2x = B2x*sxx + B2y*sxy + B2z*szx;
    real f2y = B2x*sxy + B2y*syy + B2z*syz;
    real f2z = B2x*szx + B2y*syz + B2z*szz;
    real f3x = B3x*sxx + B3y*sxy + B3z*szx;
    real f3y = B3x*sxy + B3y*syy + B3z*syz;
    real f3z = B3x*szx + B3y*syz + B3z*szz;
    real f0x = - f1x - f2x - f3x;
    real f0y = - f1y - f2y - f3y;
    real f0z = - f1z - f2z - f3z;
    ndF[0] = nd[0];   f[ 0] = f0x; f[ 1] = f0y; f[ 2] = f0z;
    ndF[1] = nd[1];   f[ 3] = f1x; f[ 4] = f1y; f[ 5] = f1z;
    ndF[2] = nd[2];   f[ 6] = f2x; f[ 7] = f2y; f[ 8] = f2z;
    ndF[3] = nd[3];   f[ 9] = f3x; f[10] = f3y; f[11] = f3z;

    return M_BIG;
}

void myElement::setUpStiffnessMatrix (real **s) {
    int ndf = 12;
    int i, j;
    for (i=0; i<ndf; i++)
        for (j=0; j<ndf; j++)
            s[i][j] = 0.;
    real vel = 1.e-6;
    Node **ndF = new Node*[4];
    real *f = new real[12];
    for (i=0; i<ndf; i++) {
        nd[0]->clrVelocities();
        nd[0]->clrForcesMoments();
        nd[1]->clrVelocities();
        nd[1]->clrForcesMoments();
        nd[2]->clrVelocities();
        nd[2]->clrForcesMoments();
        nd[3]->clrVelocities();
        nd[3]->clrForcesMoments();
        //  reset state variables
        for (j=0; j<nsv; j++)
            stv[j] = 0.;
        if (i == 0)       nd[0]->setVx(vel);
        else if (i == 1)  nd[0]->setVy(vel);
        else if (i == 2)  nd[0]->setVz(vel);
        else if (i == 3)  nd[1]->setVx(vel);
        else if (i == 4)  nd[1]->setVy(vel);
        else if (i == 5)  nd[1]->setVz(vel);
        else if (i == 6)  nd[2]->setVx(vel);
        else if (i == 7)  nd[2]->setVy(vel);
        else if (i == 8)  nd[2]->setVz(vel);
        else if (i == 9)  nd[3]->setVx(vel);
        else if (i == 10) nd[3]->setVy(vel);
        else if (i == 11) nd[3]->setVz(vel);
        calcForces(ndF, f);
        for (j=0; j<12; j++)
            s[i][j] = f[j] / vel;
    }
}

// --------------------------------------------------------------------

class ttL_myElement : public ttLst {
private:
public:
    ttL_myElement() { initList(); }
    ttL_myElement(string nm) : ttLst(nm) { initList(); }
    void initList() { }
    ~ttL_myElement() { }
    Object *createNewObject() { return new myElement(); }
    void setMaterial(Material *m) { mat = m; };
    void setNodeList(ndLst *nL) { ndL = nL; };
    void setStaticVariables();
    virtual bool processCommand(Reader *, string);
    void initialize();
    real calcFrc();
    void writeRst(RstWriter *) { }
    void readRst(RstReader *) { }
};

void ttL_myElement::setStaticVariables () {
    DeformableTet::setMaterial(mat);
    int nsv = mat->getNumberHexSV(); 
    DeformableTet::setNsv(nsv);
}

bool ttL_myElement::processCommand (Reader *rdr, string lbl) {
    if (lbl == "myKeyword") {
    } else if (ttLst::processCommand(rdr, lbl)) {
    } else
        return false;
    return true;
}

void ttL_myElement::initialize () {
    setStaticVariables();
    myElement **tt = (myElement **)ob;
    if (refSys == NULL) {
        cout << "  Reference system not specified. Using default RefSys"
            << endl;
        Lists *lists = (Lists *)lst;
        refSys = FindObj("DFLT", ReferenceSystem);
    }
    if (refSys) {
        int i;
        for (i=0; i<num; i++) 
            tt[i]->init(refSys);
    }
}

real ttL_myElement::calcFrc () {
    if (listIsNotActive()) return M_BIG;
    bool mdb = getMpiDebug();  // mpi debug
    if (mdb) console.pause("Mpi.ttL.cf.1");
    obLst::calcFrcInit();
    setStaticVariables();
    obLst::setComputeTimeStepFlag(false);
    obLst::createArrays(4, 0);
    Node **ndF = (Node **)obF;
    myElement **tt = (myElement **)lo;
    int i, ntt = nlo;
#ifdef _OPENMP
#pragma omp parallel private(i)
#pragma omp for
#endif
    for (i=0; i<ntt; i++)
        tt[i]->calcForces(ndF+4*i, f+12*i);
    if (mdb) console.pause("Mpi.ttL.cf.2");
    obLst::procForces();
    obLst::freeArrays();
    return dtm;
}
