// -----------------------------------------------------
// Important notice:
// -----------------------------------------------------
//   User Defined Object (myMaterial)
// -----------------------------------------------------
//   - use init0() in constructor to set default values to member variables
//   - MARS AUTOMATICALLY calls Object::readObjInput(), which REPEATLY calls processCommand() to process input file
//   - MARS AUTOMATICALLY calls myMaterial::init() after the input is read,
//   - use myMaterial::init() to initialize the object using the parameters read from input
//   - put the checks in myMaterial::init()

class myMaterial : public Material {
private:
    real lmb, twg, ps1, ps2, blk;
    real ymdDt, lmbDt, twgDt, ps1Dt, ps2Dt;
public:
    myMaterial(string nm) : Material(nm) { init0(); }
    void init0();
    ~myMaterial() { }
    void init();
    virtual bool processCommand(Reader *, string);
    void execAtEveryStep();
    void calcStrsHx(real *, real *) const;
    void writeRstName(RstWriter *);
    void writeRst(RstWriter *);
    void readRst(RstReader *);
};

void myMaterial::init0 () {
    type = MATER;
    // state variables used in tensorial formulations
    nHsv = 6;
    setTnsrSV(1, "XX-Stress", "stress");
    setTnsrSV(2, "YY-Stress", "stress");
    setTnsrSV(3, "ZZ-Stress", "stress");
    setTnsrSV(4, "YZ-Stress", "stress");
    setTnsrSV(5, "ZX-Stress", "stress");
    setTnsrSV(6, "XY-Stress", "stress");
    lmb = twg = ps1 = ps2 = blk = 0.;
    ymdDt = lmbDt = twgDt = ps1Dt = ps2Dt = 0.;
}

bool myMaterial::processCommand (Reader *rdr, string lbl) {
    if (lbl == "Density")
        dns = rdr->readDensity();
    else if (lbl == "YoungModulus")
        ymd = rdr->readStress();
    else if (lbl == "PoissonsRatio") 
        psn = rdr->getFloat();
    else if (Material::processCommand(rdr, lbl)) {
    } else 
        return false;
    return true;
}

void myMaterial::init () {
    lmb = ymd * psn / ((1.+psn)*(1.-2.*psn));
    twg = ymd / (1.+psn);
    blk = lmb + twg/3.;
    ps1 = ymd / (1.-psn*psn);
    ps2 = ps1 * psn;
    ssp = sqrt(blk/dns);
    ps1Dt = ps2Dt = 0.;
}

void myMaterial::execAtEveryStep () {
    real dt = prm->getTimeStep();
    ymdDt = ymd * dt;
    twgDt = twg * dt;
    lmbDt = lmb * dt;
    ps1Dt = ps1 * dt;
    ps2Dt = ps2 * dt;
}

void myMaterial::calcStrsHx (real *hxv, real *stv) const {
    real hin = lmbDt * (hxv[0] + hxv[1] + hxv[2]);
    stv[0] += hin + twgDt*hxv[0];
    stv[1] += hin + twgDt*hxv[1];
    stv[2] += hin + twgDt*hxv[2];
    stv[3] += twgDt*hxv[3];
    stv[4] += twgDt*hxv[4];
    stv[5] += twgDt*hxv[5];
}

void myMaterial::writeRstName (RstWriter *rst) {
    rst->write(name);
    rst->write("ela");
}

void myMaterial::writeRst (RstWriter *rst) {
    rst->write(name);
    rst->write('a');
    rst->write(dns);
    rst->write(ymd);
    rst->write(psn);
    rst->writeMark();
}

void myMaterial::readRst (RstReader *rst) {
    rst->readByt();
    rst->read(dns);
    rst->read(ymd);
    rst->read(psn);
    rst->readMark("myMaterial");
}
